
/* 
    This program uses a pure resistive load and calculates Ah and current. It does not measure the voltage drop
    across the load to measure current and resistance accurately therefore will only give approximate Ah rating however
    it will give a total time of how long the battery lasts with a load on it in comparison with a known good battery.

    Version 2.01 - improved display layout. Fixed bug with top line of display showing wrong info when test done. Added RTC temperature display
	  Hardware change to use precision 4.096V reference instead of TL431.
	  Removed voltage averaging as it slows down the CPU and does not improve accuracy anyway.
            
    Load = 2.5 ohms for 60Ah, 5 ohms for 30Ah.
   
    Pin / port listings (Arduino Nano):-

    0 - COM1 TX (programming / diagnostics)
    1 - COM1 RX (programming / diagnostics)
    2 - LCD Data
    3 - LCD Data
    4 - LCD Data
    5 - LCD Data
    6 - 30/60Ah select (input)
    7 - Relay to select load bank(s) out
    8 - Relay out for load bank and fan
    9 - COM2 TX (Printer)
    10 - COM2 RX (Printer)
    11 - LCD Control
    12 - LCD Control
    13 - Test complete LED
    A0 - Vsense
    A1 - Tsense (reserved, not implemented in this version)
    A2
    A3
    A4 - SDA (for RTC)
    A5 - SCL (for RTC)
    
       
    ***Calibration***
    
    The voltage divider factor is calculated by dividing the first voltage by the second voltage or:
    dividing factor = input voltage ÷ output voltage
    For example, if the first or input voltage measured is 10.02V and the second or output voltage is 0.9V, then the division factor is:
    10.02 ÷ 0.9 = 11.133
    Voltages should be measured with a good calibrated DVM for best accuracy and should be recalibrated once every year.
      
*/

#include <LiquidCrystal.h>
#include <SoftwareSerial.h>
#include "thermalprinter.h"
#include "eepromanything.h"
#include <Wire.h>
#include "RTClib.h"
#include <avr/pgmspace.h> 
#include <avr/io.h>

const float  adacRef = 4.096; 
const int numReadings = 6;
int readings[numReadings];      // the readings from the analog input
int readIndex = 0;              // the index of the current reading
int total = 0;                  // the running total
char tVal[6];

// The integrator works by accumulating the sampled voltage values from the start until
//   hitting a pre-determined low-voltage threshold. Since they are sampled at a 
//   known period, the number of samples taken cancels out and the accumulation of all samples 
//   simply needs to be scaled by a factor
//
//   Let's make some definitions to show the derivation of how this is so:
//   I = current 
//   V = load voltage
//   Rl = load resistance
//   samples = number of voltage samples made
//   sensor = value read out of analog A->D
//   rate = number of samples taken per second
//   
//                                        ave(V)
//   capacity = ∫I ~= ave(I) * time = ----------- * time
//                                          Rl
//
//   Where:
//                ∑(V)     ∑(sensor) * adacRef/1024
//     ave(V) = -------  = ---------------------------------
//              samples              samples
//
//      time  =  samples/rate 
//
//   Thus:
//               ∑(sensor) * adacRef        samples
//   capacity = ----------------------------- * ----------
//                samples * 1024 * Rl            rate
//
//               ∑(sensor) * adacRef 
//   capacity = ------------------------------ = ∑(sensor) * quanta 
//                  1024 * Rl * rate
//
//
//                adacRef
//   quanta =  --------------------  
//               1024 * Rl * rate
//
//   rate in this sketch is 1 Hz (1/sec).  
//   Thus, the units of quanta are: volt*seconds/(Ω*samples) = amp*seconds per sample
//
//   What we really want are mA*Hours, so, scale by 1hr/3600seconds and 1000mA/A, or simply 1/3.6
//
//   So, finally:
//   quanta = adacRef/(Rl*1024*3.6)  in units of mA*hours per sample

const double quanta30 = 0.00021787; // 4.096/(5.1*1024*3.6) // 18800
const double quanta60 = 0.00042737; // 4.096/(2.6*1024*3.6) // 9584
double quanta;

// Define low voltage threshold where any remaining capacity is "unusable"
//        lowThreshold = 21 / VDIV  is 1.88 volts so 1.88*1024/adacRef = 470 + offset of 1 to get exactly 21V
const int lowThreshold = 471; // (21V)

int sensorPin          = A0;    // select the analog input pin for the voltage measurement
int ledPin             = 13;   // select the pin for the test complete LED
int loadselectPin      = 7;    // Pin load bank select relay
int fetGatePin         = 8;    // select the pin for the FET
int printStatus        = 0;    // for printer init
int ahSelectsw         = 6;    // Pin for the selector switch
int sensorValue        = 0;    // unscaled sensor output 
int val                = 0;    // variable to store the value coming from the temp sensor
float voltage          = 0;    // measured voltage
float loadcurrent      = 0;    // calculated load current
double mAH             = 0;    // calculated mAh
long accumulator       = 0;    // sum of all unscaled sensor values sampled
int epoch              = 0;    // seconds since battery was inserted 
int lowVolts           = 0;    // de-bounce the low voltage threshold
bool done           = false;// Voltage has dropped below threshold
bool batteryIn      = false;// Battery present
bool printed        = false;// to determine if printer has printed or not
bool numwritten     = false;// to determine if test number is written to EEPROM
long testnum;
long x;
#define DS3231_I2C_ADDR             0x68 // Used to get temperature from internal DS3231 sensor
#define DS3231_TEMPERATURE_ADDR     0x11 // Used to get temperature from internal DS3231 sensor
#define R1 2.6; // actual value at 25 degrees C = 5.1 ohms each resistor. At 90 degrees C 5.2 ohms.
#define R2 5.1;
#define rxPin 10 // used for software serial to printer
#define txPin 9 // used for software serial to printer
#define VDIV 11.1365 // ADAC is only 10 bit so voltage readings are in 5mv steps. 11.1572 calculated value, may need to fine tune this value. 
#define VCAL 1.167 // as it is a resistive load Ah rating will be slightly less than actual capacity, this compensates for that. Gives around 5% accuracy. 

const char *daysOfTheWeek[] = {"SUN", "MON", "TUE", "WED", "THU", "FRI", "SAT"};
const char *monthsOfyear[]  = {"JAN", "FEB", "MAR", "APR", "MAY", "JUN", "JUL", "AUG", "SEP", "OCT", "NOV", "DEC"};

LiquidCrystal lcd(12, 11,  5,  4,  3,  2); //  LCD Pins  --  RS, EN, D4, D5, D6, D7
Epson TM88 = Epson(rxPin, txPin); // init the Printer with Output-Pin

RTC_DS3231 rtc; // comment this back in for hardware RTC and comment out above line

const int DS1307 = 0x68; // Address of DS1307 see data sheets
 
// Initializes all values: 
byte second = 0;
byte minute = 0;
byte hour = 0;
byte weekday = 0;
byte monthday = 0;
byte month = 0;
byte year = 0;



byte decToBcd(byte val) 
{
  return ((val/10*16) + (val%10));
}
byte bcdToDec(byte val) 
{
  return ((val/16*10) + (val%16));
}

// RUN ONCE AT POWER ON

void setup() 
{
  
  pinMode(ledPin,     OUTPUT);  
  pinMode(fetGatePin, OUTPUT);  
  pinMode(loadselectPin, OUTPUT);
  pinMode(ahSelectsw, INPUT_PULLUP); // set to input and turn on internal 10K pullup resistor
  
  //rtc.adjust(DateTime(F(__DATE__), F(__TIME__))); // for hardware RTC
  analogReference(EXTERNAL); // set analog reference to external
  Serial.begin(9600);
  TM88.start();
  lcd.begin(20, 2);
  rtc.begin();
  digitalWrite(ledPin, HIGH); // LED will be connected to LCD backlight
  lcd.print("Starting...");
  
 /*   Removed in version 2.01
 for (int thisReading = 0; thisReading < numReadings; thisReading++) 
  {
   readings[thisReading] = 0;
  }
*/
serialMenu();
lcd.clear();
startUp();
  
}

// MAIN PROGRAM LOOP 

void loop() {
/*    Removed in version 2.01
  total = total - readings[readIndex]; // take a number of samples from the ADAC and add them up
  readings[readIndex] = analogRead(sensorPin);
  total = total + readings[readIndex];
  readIndex = readIndex + 1;

  if (readIndex >= numReadings) 
    {
      readIndex = 0;
    }

  sensorValue = total / numReadings; // average the ADAC samples so they can be used to calculate voltage later
*/    
  if (digitalRead (ahSelectsw) == HIGH) // load select INPUT is active LOW so when 60Ah mode is selected switch will pull pin to GND.
  {
  quanta = quanta30; 
  loadcurrent = voltage/R2; 
  digitalWrite(loadselectPin, LOW); // disconnects two of the resistors so that only remaining two are in parallel.
  }
  else 
  {
  quanta = quanta60; 
  loadcurrent = voltage/R1;
  digitalWrite(loadselectPin, HIGH); // switches on all four load resistors for 60Ah batteries so total parallel resistance is 2.5 ohms
  }
  
  sensorValue = analogRead(sensorPin);    

  if (!batteryIn && printed == false && (sensorValue > 560)) {  //Only start test if battery is above 25V
    // Initialize upon the insertion of a "fresh" battery 
    batteryIn   = true;
    done        = false;
    printed     = false;
    epoch       = 0;
    accumulator = 0;
    lowVolts    = 0;
    printStart();
    digitalWrite(fetGatePin, HIGH);
    // clear out when LCD when starting over
    lcd.clear();
  }
  else if (batteryIn && done && (sensorValue <= 20)) {
    // consider the battery disconnected only after finishing the last measurement to 
    // debounce a glitchy contact
    batteryIn=false;
  } 
  else if (batteryIn && !done) {
    // Running state during discharge state
 
    voltage = sensorValue*adacRef/1024.0 *VDIV;
    //loadcurrent = voltage/R1; // this is now declared in setup depending on load selected
  
    // print voltage
    lcd.setCursor(0, 0);
    if (voltage <10)
    lcd.print("0");
    lcd.print(voltage);
    lcd.print("V");
    lcd.setCursor(8, 0);
    lcd.print(loadcurrent);
    lcd.print(F("A "));

    float ds3231temp = DS3231_get_treg();
    dtostrf(ds3231temp-2,4,1,tVal);    // convert temperature value to xx.x format       was 4,2,tVal
    
    // print temperature here pos 15,0 to 1 decimal place

    lcd.setCursor(15,0);
    lcd.print(tVal);
    lcd.print("C");
      
    if (!done) {
      accumulator += sensorValue;
      // print mAH
      mAH = accumulator*quanta;
      lcd.setCursor(0,1);
      if  (mAH /100 *VCAL < 10) 
      lcd.print("0");
      lcd.print(double(mAH /100 *VCAL));
      lcd.print("Ah ");
      //if  (mAH < 10000) // shift display if more than 10,000mAh (10Ah)
      lcd.setCursor(8,1); 

    int chours, cminutes, cseconds;
      chours = epoch/60 / 60;
      cminutes = epoch /60 % 60;
      cseconds = epoch %60;
      if (chours <= 9)
      {
      lcd.print("0");
      }
      lcd.print(chours);
      lcd.print(":");
      if (cminutes <= 9)
      {
       lcd.print("0");
      }
       lcd.print(cminutes);
       lcd.print(":");
      if (cseconds <= 9)
      {
       lcd.print("0");
      }
       lcd.print(cseconds);          
    }
  
    if (!done){
  
      // lowVolts requires 10 seconds in the last 20 before being done
      if (sensorValue < lowThreshold) { lowVolts++;}
      else if (lowVolts > 0)          { lowVolts--;}  
  
      // If it's below threshold for 10 of 20 samples, bail out
      if (lowVolts > 10) 
      {
        done=true;
        digitalWrite(fetGatePin, LOW);
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print(F("Capacity "));
        lcd.print(double(mAH /100 *VCAL)); // display result in amp hours
        lcd.print(" Ah ");
        lcd.setCursor(0, 1);
        lcd.print(F("Runtime "));
        
        int chours, cminutes, cseconds;
        chours = epoch/60 / 60;
        cminutes = epoch /60 % 60;
        if (chours <= 9)
        {
         lcd.print("0");
        }
       lcd.print(chours);
       lcd.print("h ");
       if (cminutes <= 9)
        {
         lcd.print("0");
        }
       lcd.print(cminutes);
       lcd.print("m");        
    
      }
    }

    epoch++;
    
  } // batteryIn -- main routine

  if (done) {
    // When done, flash the LED to get attention
    PORTB |= (1<<PORTB5); // set pin 13 high
	//digitalWrite(ledPin, HIGH);
    delay(500);
	PORTB &= ~(1<<PORTB5); // set pin 13 low
    //digitalWrite(ledPin, LOW);
    delay(499);

// increment test number and write to EEPROM
       if (numwritten == false)
       {
       EEPROM_readAnything(0, testnum);
       x = testnum; 
       testnum = ++x;
       EEPROM_writeAnything(0, testnum); 
       numwritten = true;
       }
    
// print out test results

       if (printed == false)

       {
         printEnd();
       }
    // do something here if test restarts when battery voltage rebounds after load disconnect
  } else 
  
  {
    // Since the processing takes some time prior to the delay, we'll assume 1mS 
    delay(999); // really should improve this as we are using an RTC chip in this version. It works though so maybe in future version.
  }

// DEBUG
/*
  Serial.print("\load current in amps = " ); 
  Serial.print(loadcurrent); 
  Serial.println( );
  
  Serial.print("\nsensor = " );                       
  Serial.print(sensorValue);      
  Serial.print("\t lowvolts = " );                       
  Serial.print(lowVolts);      
*/
  
}

// FUNCTIONS

float DS3231_get_treg()
{
    short rv;  
    
    Wire.beginTransmission(DS3231_I2C_ADDR);
    Wire.write(DS3231_TEMPERATURE_ADDR);
    Wire.endTransmission();

    Wire.requestFrom(DS3231_I2C_ADDR, 2);
    rv = Wire.read() << 8;
    rv |= Wire.read();
  return(rv/256.);
}

void startUp()

{
  DateTime now = rtc.now();
  lcd.print(F("custom txt here"));
  lcd.setCursor(0, 1);
  lcd.print(F("S/W ver 2.01"));
  delay(2000);
  lcd.setCursor(0, 0);
  lcd.clear();
  lcd.print(F("LiFePo4 Lithium")); 
  lcd.setCursor(0, 1);
  lcd.print(F("Battery Tester"));
  delay(2000);
  lcd.setCursor(0, 0);
  lcd.clear();
  if(now.hour()<10) lcd.print(0, DEC);
  lcd.print(now.hour(), DEC);
  lcd.print(':');
  if(now.minute()<10) lcd.print(0, DEC); 
  lcd.print(now.minute(), DEC); 
  lcd.setCursor(0, 1);
  lcd.print(daysOfTheWeek[now.dayOfTheWeek()]);
  lcd.print('-');
  lcd.print(now.day(), DEC);
  lcd.print('-');
  lcd.print(monthsOfyear[now.month()-1]);
  lcd.print('-');
  lcd.print(now.year(), DEC);
  delay(4000);
  lcd.clear();  
  lcd.setCursor(0, 0);
  Serial.println(F("LIFEPO4 BATTERY TESTER"));

  
    if (digitalRead (ahSelectsw) == HIGH)
  {
  lcd.print(F("30Ah mode"));
  Serial.println(F("30Ah mode"));
  }

else 
  {
  lcd.print(F("60Ah Mode"));
  Serial.println(F("60Ah mode"));
  }

  delay (4000);
  lcd.clear();  
  lcd.setCursor(0, 0);

  // Print a helpful start-up message to the LCD.
  lcd.print(F("Connect battery"));
  lcd.setCursor(0, 1);
  lcd.print(F("to start the test"));
  
}


void printStart()

{
  voltage = sensorValue*adacRef/1024.0 *VDIV;
  DateTime now = rtc.now();
  TM88.defaultLineSpacing();
  TM88.justifyCenter();
  TM88.println(F("LiFePo4 Battery Tester"));
  TM88.println(F("S/W Ver 2.01 Adrian Smith"));
  TM88.feed(2);
  TM88.justifyLeft();
  EEPROM_readAnything(0, testnum);
  TM88.print(F("Test number "));
  Serial.print(F("Test Number "));
  Serial.print(testnum);
  TM88.print(testnum);
  Serial.println(" ");
  TM88.feed(1);
  TM88.println(F("Test started at"));
  Serial.println(F("Test started at"));
  if(now.hour()<10) TM88.print(0, DEC);
  TM88.print(now.hour(), DEC);
  TM88.print(":"); 
  if(now.minute()<10) TM88.print(0, DEC);
  TM88.print(now.minute(), DEC);
  TM88.println(" ");
  TM88.print(daysOfTheWeek[now.dayOfTheWeek()]);
  TM88.print("-");
  TM88.print(now.day(), DEC);
  TM88.print("-");
  TM88.print(monthsOfyear[now.month()-1]);
  TM88.print("-");
  TM88.print(now.year(), DEC);
  TM88.feed(1);
  TM88.println(F("Battery start voltage"));
  TM88.print(voltage);
  TM88.feed(4); 
  Serial.print((now.hour(), DEC));
  Serial.print(":"); 
  if(now.minute()<10) Serial.print(0, DEC);
  Serial.print(now.minute(), DEC);
  Serial.println(" ");
  Serial.print(daysOfTheWeek[now.dayOfTheWeek()]);
  Serial.print("-");
  Serial.print(now.day(), DEC);
  Serial.print("-");
  Serial.print(monthsOfyear[now.month()-1]);
  Serial.print("-");
  Serial.print(now.year(), DEC);
}

void printEnd()

{
   DateTime now = rtc.now();
   TM88.justifyLeft();
   TM88.defaultLineSpacing();  
   TM88.println(F("Test finished at"));
   Serial.println(" ");
   Serial.println(" ");
   Serial.println(F("Test finished at"));
   Serial.print(now.hour(), DEC);
   Serial.print(":"); 
   if(now.minute()<10) Serial.print(0, DEC);
   Serial.print(now.minute(), DEC);
   Serial.println(" ");
   Serial.print(daysOfTheWeek[now.dayOfTheWeek()]);
   Serial.print("-");
   Serial.print(now.day(), DEC);
   Serial.print("-");
   Serial.print(monthsOfyear[now.month()-1]);
   Serial.print("-");
   Serial.print(now.year(), DEC);
   if(now.hour()<10) TM88.print(0, DEC);
   TM88.print(now.hour(), DEC);
   TM88.print(":"); 
   if(now.minute()<10) TM88.print(0, DEC);
   TM88.print(now.minute(), DEC);
   TM88.println(" ");
   TM88.print(daysOfTheWeek[now.dayOfTheWeek()]);
   TM88.print("-");
   TM88.print(now.day(), DEC);
   TM88.print("-");
   TM88.print(monthsOfyear[now.month()-1]);
   TM88.print("-");
   TM88.print(now.year(), DEC);
   TM88.feed(2); 
   TM88.println(F("Battery end voltage (offload)"));
   Serial.println(" ");
   Serial.println(F("Battery end voltage (offload)"));
   TM88.print(voltage);
   Serial.print(voltage);
   TM88.feed(1);
   TM88.println(F("Battery measured capacity "));
   TM88.print(double(mAH /100 *VCAL)); // display result in amp hours
   TM88.print(F(" Ah +/- 5% @ 25 degrees C"));
   TM88.feed(1);
      
   int chours, cminutes;
   chours = epoch/60 / 60;
   cminutes = epoch /60 % 60;
   TM88.println(F("Total runtime (hours, mins)"));
    if (chours <= 9)
   {
    TM88.print("0");
   }
   TM88.print(chours);
   TM88.print(":");
   if (cminutes <= 9)
   {
    TM88.print("0");
   }
   TM88.print(cminutes);
   TM88.feed(5);
   TM88.cut();  
   printed = true; // prevents from printing over and over
   
  }

void serialMenu()
{
  delay(100); // This delay allows the MCU to read the current date and time.
 
  Serial.print(F("The current date and time is: "));
  printTime();
  Serial.println(F("Please change to newline in the settings on the lower right of the Serial Monitor"));
  Serial.println(F("Would you like to set the date and time now? Y/N"));
  
  if (Serial.available()>=0) delay(5000); // wait 5 seconds for user input otherwise continue boot
  if (Serial.read() == 'y' || Serial.read() == 'Y')
 
  // This set of functions allows the user to change the date and time
  {
    Serial.read();
    setTime();
    Serial.print(F("The current date and time is now: "));
    printTime();
  }
 
 
  Serial.println(F("Thank you."));
}

void printTime() {
  char buffer[3];
  const char* AMPM = 0;
  readTime();
  Serial.print(daysOfTheWeek[weekday-1]);
  Serial.print(" ");
  Serial.print(monthsOfyear[month-1]);
  Serial.print(" ");
  Serial.print(monthday);
  Serial.print(", 20");
  Serial.print(year);
  Serial.print(" ");
  if (hour > 12) {
    hour -= 12;
    AMPM = " PM";
  }
  else AMPM = " AM";
  Serial.print(hour);
  Serial.print(":");
  sprintf(buffer, "%02d", minute);
  Serial.print(buffer);
  Serial.println(AMPM);
}


void readTime() {
  Wire.beginTransmission(DS1307);
  Wire.write(byte(0));
  Wire.endTransmission();
  Wire.requestFrom(DS1307, 7);
  second = bcdToDec(Wire.read());
  minute = bcdToDec(Wire.read());
  hour = bcdToDec(Wire.read());
  weekday = bcdToDec(Wire.read());
  monthday = bcdToDec(Wire.read());
  month = bcdToDec(Wire.read());
  year = bcdToDec(Wire.read());
}

void setTime() {
  Serial.print(F("Please enter the current year, 00-99. - "));
  year = readByte();
  Serial.println(year);
  Serial.print(F("Please enter the current month, 1-12. - "));
  month = readByte();
  Serial.println(monthsOfyear[month-1]);
  Serial.print(F("Please enter the current day of the month, 1-31. - "));
  monthday = readByte();
  Serial.println(monthday);
  Serial.println(F("Please enter the current day of the week, 1-7."));
  Serial.print(F("1 Sun | 2 Mon | 3 Tues | 4 Weds | 5 Thu | 6 Fri | 7 Sat - "));
  weekday = readByte();
  Serial.println(daysOfTheWeek[weekday-1]);
  Serial.print(F("Please enter the current hour in 24hr format, 0-23. - "));
  hour = readByte();
  Serial.println(hour);
  Serial.print(F("Please enter the current minute, 0-59. - "));
  minute = readByte();
  Serial.println(minute);
  second = 0;
  Serial.println(F("The data has been entered."));
 
  // The following codes transmits the data to the RTC
  Wire.beginTransmission(DS1307);
  Wire.write(byte(0));
  Wire.write(decToBcd(second));
  Wire.write(decToBcd(minute));
  Wire.write(decToBcd(hour));
  Wire.write(decToBcd(weekday));
  Wire.write(decToBcd(monthday));
  Wire.write(decToBcd(month));
  Wire.write(decToBcd(year));
  Wire.write(byte(0));
  Wire.endTransmission();
  // Ends transmission of data
}


byte readByte() {
  while (!Serial.available()) delay(10);
  byte reading = 0;
  byte incomingByte = Serial.read();
  while (incomingByte != '\n') {
    if (incomingByte >= '0' && incomingByte <= '9')
      reading = reading * 10 + (incomingByte - '0');
    else;
    incomingByte = Serial.read();
  }
  Serial.flush();
  return reading;
}
